# type: ignore
import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Optional

from pydantic import Field, Secret, ValidationError
from pydantic.functional_validators import BeforeValidator

from unstructured_ingest.embed.interfaces import (
    AsyncBaseEmbeddingEncoder,
    BaseEmbeddingEncoder,
    EmbeddingConfig,
)
from unstructured_ingest.errors_v2 import UserAuthError, is_internal_error
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from vertexai.language_models import TextEmbeddingModel


def conform_string_to_dict(value: Any) -> dict:
    if isinstance(value, dict):
        return value
    if isinstance(value, str):
        return json.loads(value)
    raise ValidationError(f"Input could not be mapped to a valid dict: {value}")


ApiKeyType = Secret[Annotated[dict, BeforeValidator(conform_string_to_dict)]]


class VertexAIEmbeddingConfig(EmbeddingConfig):
    api_key: ApiKeyType = Field(description="API key for Vertex AI")
    embedder_model_name: Optional[str] = Field(
        default="text-embedding-005", alias="model_name", description="Vertex AI model name"
    )

    def wrap_error(self, e: Exception) -> Exception:
        if is_internal_error(e=e):
            return e
        from google.auth.exceptions import GoogleAuthError

        if isinstance(e, GoogleAuthError):
            return UserAuthError(e)
        return e

    def register_application_credentials(self):
        # TODO look into passing credentials in directly, rather than via env var and tmp file
        application_credentials_path = Path("/tmp") / "google-vertex-app-credentials.json"
        with application_credentials_path.open("w+") as credentials_file:
            json.dump(self.api_key.get_secret_value(), credentials_file)
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = str(application_credentials_path)

    @requires_dependencies(
        ["vertexai"],
        extras="vertexai",
    )
    def get_client(self) -> "TextEmbeddingModel":
        """Creates a VertexAI python client to embed elements."""
        from vertexai.language_models import TextEmbeddingModel

        self.register_application_credentials()
        return TextEmbeddingModel.from_pretrained(self.embedder_model_name)


@dataclass
class VertexAIEmbeddingEncoder(BaseEmbeddingEncoder):
    config: VertexAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "TextEmbeddingModel":
        return self.config.get_client()

    @requires_dependencies(
        ["vertexai"],
        extras="embed-vertexai",
    )
    def embed_batch(self, client: "TextEmbeddingModel", batch: list[str]) -> list[list[float]]:
        from vertexai.language_models import TextEmbeddingInput

        inputs = [TextEmbeddingInput(text=text) for text in batch]
        response = client.get_embeddings(inputs)
        return [e.values for e in response]


@dataclass
class AsyncVertexAIEmbeddingEncoder(AsyncBaseEmbeddingEncoder):
    config: VertexAIEmbeddingConfig

    def wrap_error(self, e: Exception) -> Exception:
        return self.config.wrap_error(e=e)

    def get_client(self) -> "TextEmbeddingModel":
        return self.config.get_client()

    @requires_dependencies(
        ["vertexai"],
        extras="embed-vertexai",
    )
    async def embed_batch(self, client: Any, batch: list[str]) -> list[list[float]]:
        from vertexai.language_models import TextEmbeddingInput

        inputs = [TextEmbeddingInput(text=text) for text in batch]
        response = await client.get_embeddings_async(inputs)
        return [e.values for e in response]



# https://developers.notion.com/reference/parent-object
from dataclasses import dataclass

from unstructured_ingest.processes.connectors.notion.interfaces import FromJSONMixin


# https://developers.notion.com/reference/parent-object#database-parent
@dataclass
class DatabaseParent(FromJSONMixin):
    database_id: str
    type: str = "database_id"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(database_id=data["database_id"])


# https://developers.notion.com/reference/parent-object#page-parent
@dataclass
class PageParent(FromJSONMixin):
    page_id: str
    type: str = "page_id"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(page_id=data["page_id"])


# https://developers.notion.com/reference/parent-object#workspace-parent
@dataclass
class WorkspaceParent(FromJSONMixin):
    type: str = "workspace"
    workspace: bool = True

    @classmethod
    def from_dict(cls, data: dict):
        return cls()


# https://developers.notion.com/reference/parent-object#block-parent
@dataclass
class BlockParent(FromJSONMixin):
    block_id: str
    type: str = "block_id"

    @classmethod
    def from_dict(cls, data: dict):
        return cls(block_id=data["block_id"])


@dataclass
class Parent(FromJSONMixin):
    block_id: str
    type: str = "block_id"

    @classmethod
    def from_dict(cls, data: dict):
        t = data["type"]
        if t == "database_id":
            return DatabaseParent.from_dict(data)
        elif t == "page_id":
            return PageParent.from_dict(data)
        elif t == "workspace":
            return WorkspaceParent.from_dict(data)
        elif t == "block_id":
            return BlockParent.from_dict(data)



from collections import OrderedDict
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Generator, List, Optional, Tuple

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import SourceConnectionError
from unstructured_ingest.errors_v2 import UserAuthError, UserError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    download_responses,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    SourceRegistryEntry,
)
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.html import HtmlMixin
from unstructured_ingest.utils.string_and_date_utils import fix_unescaped_unicode

if TYPE_CHECKING:
    from atlassian import Confluence
    from bs4 import BeautifulSoup
    from bs4.element import Tag

CONNECTOR_TYPE = "confluence"


class ConfluenceAccessConfig(AccessConfig):
    password: Optional[str] = Field(
        description="Confluence password",
        default=None,
    )
    api_token: Optional[str] = Field(
        description="Confluence Cloud API token",
        default=None,
    )
    token: Optional[str] = Field(
        description="Confluence Personal Access Token",
        default=None,
    )


class ConfluenceConnectionConfig(ConnectionConfig):
    url: str = Field(description="URL of the Confluence instance")
    username: Optional[str] = Field(
        description="Username or email for authentication",
        default=None,
    )
    cloud: bool = Field(description="Authenticate to Confluence Cloud", default=False)
    access_config: Secret[ConfluenceAccessConfig] = Field(
        description="Access configuration for Confluence"
    )

    def model_post_init(self, __context):
        access_configs = self.access_config.get_secret_value()
        if access_configs.password and access_configs.api_token:
            raise ValueError(
                "both password and api_token provided, only one allowed, "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        basic_auth = bool(self.username and (access_configs.password or access_configs.api_token))
        pat_auth = access_configs.token
        if self.cloud and not basic_auth:
            raise ValueError(
                "cloud authentication requires username and API token (--password), "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        if basic_auth and pat_auth:
            raise ValueError(
                "both password and token provided, only one allowed, "
                "see: https://atlassian-python-api.readthedocs.io/"
            )
        if not (basic_auth or pat_auth):
            raise ValueError(
                "no form of auth provided, see: https://atlassian-python-api.readthedocs.io/"
            )

    def password_or_api_token(self) -> str:
        # Confluence takes either password or API token under the same field: password
        # This ambiguity led to confusion, so we are making it specific what you are passing in
        access_configs = self.access_config.get_secret_value()
        if access_configs.password:
            return access_configs.password
        return access_configs.api_token

    @requires_dependencies(["atlassian"], extras="confluence")
    @contextmanager
    def get_client(self) -> Generator["Confluence", None, None]:
        from atlassian import Confluence

        access_configs = self.access_config.get_secret_value()
        with Confluence(
            url=self.url,
            username=self.username,
            password=self.password_or_api_token(),
            token=access_configs.token,
            cloud=self.cloud,
        ) as client:
            yield client


class ConfluenceIndexerConfig(IndexerConfig):
    max_num_of_spaces: int = Field(500, description="Maximum number of spaces to index")
    max_num_of_docs_from_each_space: int = Field(
        100, description="Maximum number of documents to fetch from each space"
    )
    spaces: Optional[List[str]] = Field(None, description="List of specific space keys to index")


@dataclass
class ConfluenceIndexer(Indexer):
    connection_config: ConfluenceConnectionConfig
    index_config: ConfluenceIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def precheck(self) -> bool:
        try:
            self.connection_config.get_client()
        except Exception as e:
            logger.exception(f"Failed to connect to Confluence: {e}")
            raise UserAuthError(f"Failed to connect to Confluence: {e}")

        with self.connection_config.get_client() as client:
            # opportunistically check the first space in list of all spaces
            try:
                client.get_all_spaces(limit=1)
            except Exception as e:
                logger.exception(f"Failed to connect to find any Confluence space: {e}")
                raise UserError(f"Failed to connect to find any Confluence space: {e}")

            logger.info("Connection to Confluence successful.")

            # If specific spaces are provided, check if we can access them
            errors = []

            if self.index_config.spaces:
                for space_key in self.index_config.spaces:
                    try:
                        client.get_space(space_key)
                    except Exception as e:
                        logger.exception(f"Failed to connect to Confluence: {e}")
                        errors.append(f"Failed to connect to '{space_key}' space, cause: '{e}'")

            if errors:
                raise UserError("\n".join(errors))

        return True

    def _get_space_ids_and_keys(self) -> List[Tuple[str, int]]:
        """
        Get a list of space IDs and keys from Confluence.

        Example space ID (numerical): 98503
        Example space key (str): "SD"
        """
        spaces = self.index_config.spaces
        if spaces:
            with self.connection_config.get_client() as client:
                space_ids_and_keys = []
                for space_key in spaces:
                    space = client.get_space(space_key)
                    space_ids_and_keys.append((space_key, space["id"]))
                return space_ids_and_keys
        else:
            with self.connection_config.get_client() as client:
                all_spaces = client.get_all_spaces(limit=self.index_config.max_num_of_spaces)
            space_ids_and_keys = [(space["key"], space["id"]) for space in all_spaces["results"]]
            return space_ids_and_keys

    def _get_docs_ids_within_one_space(self, space_key: str) -> List[dict]:
        with self.connection_config.get_client() as client:
            pages = client.get_all_pages_from_space(
                space=space_key,
                start=0,
                expand=None,
                content_type="page",  # blogpost and comment types not currently supported
                status=None,
            )
        # Limit the number of documents to max_num_of_docs_from_each_space
        # Note: this is needed because the limit field in client.get_all_pages_from_space does
        # not seem to work as expected
        limited_pages = pages[: self.index_config.max_num_of_docs_from_each_space]
        doc_ids = [{"space_id": space_key, "doc_id": page["id"]} for page in limited_pages]
        return doc_ids

    def run(self) -> Generator[FileData, None, None]:
        from time import time

        space_ids_and_keys = self._get_space_ids_and_keys()
        for space_key, space_id in space_ids_and_keys:
            doc_ids = self._get_docs_ids_within_one_space(space_key)
            for doc in doc_ids:
                doc_id = doc["doc_id"]
                # Build metadata
                metadata = FileDataSourceMetadata(
                    date_processed=str(time()),
                    url=f"{self.connection_config.url}/pages/{doc_id}",
                    record_locator={
                        "space_id": space_key,
                        "document_id": doc_id,
                    },
                )
                additional_metadata = {
                    "space_key": space_key,
                    "space_id": space_id,  # diff from record_locator space_id (which is space_key)
                    "document_id": doc_id,
                }

                # Construct relative path and filename
                filename = f"{doc_id}.html"
                relative_path = str(Path(space_key) / filename)

                source_identifiers = SourceIdentifiers(
                    filename=filename,
                    fullpath=relative_path,
                    rel_path=relative_path,
                )

                file_data = FileData(
                    identifier=doc_id,
                    connector_type=self.connector_type,
                    metadata=metadata,
                    additional_metadata=additional_metadata,
                    source_identifiers=source_identifiers,
                    display_name=source_identifiers.fullpath,
                )
                yield file_data


class ConfluenceDownloaderConfig(HtmlMixin, DownloaderConfig):
    max_num_metadata_permissions: int = Field(
        250, description="Approximate maximum number of permissions included in metadata"
    )

    @requires_dependencies(["bs4"])
    def _find_hyperlink_tags(self, html_soup: "BeautifulSoup") -> list["Tag"]:
        from bs4.element import Tag

        return [
            element
            for element in html_soup.find_all(
                "a",
                attrs={
                    "class": "confluence-embedded-file",
                    "data-linked-resource-type": "attachment",
                    "href": True,
                },
            )
            if isinstance(element, Tag)
        ]


@dataclass
class ConfluenceDownloader(Downloader):
    connection_config: ConfluenceConnectionConfig
    download_config: ConfluenceDownloaderConfig = field(default_factory=ConfluenceDownloaderConfig)
    connector_type: str = CONNECTOR_TYPE
    _permissions_cache: dict = field(default_factory=OrderedDict)
    _permissions_cache_max_size: int = 5

    def download_embedded_files(
        self, session, html: str, current_file_data: FileData
    ) -> list[DownloadResponse]:
        if not self.download_config.extract_files:
            return []
        url = current_file_data.metadata.url
        if url is None:
            logger.warning(
                f"""Missing URL for file: {current_file_data.source_identifiers.filename}.
                Skipping file extraction."""
            )
            return []
        filepath = current_file_data.source_identifiers.relative_path
        download_path = Path(self.download_dir) / filepath
        download_dir = download_path.with_suffix("")
        return self.download_config.extract_embedded_files(
            url=url,
            download_dir=download_dir,
            original_filedata=current_file_data,
            html=html,
            session=session,
        )

    def parse_permissions(self, doc_permissions: dict, space_permissions: list) -> dict[str, dict]:
        """
        Parses document and space permissions to determine final user/group roles.

        :param doc_permissions: dict containing document-level restrictions
        - doc_permissions type in Confluence: ContentRestrictionArray
        :param space_permissions: list of space-level permission assignments
        - space_permissions type in Confluence: list of SpacePermissionAssignment
        :return: dict with operation as keys and each maps to dict with "users" and "groups"

        Get document permissions. If they exist, they will override space level permissions.
        Otherwise, apply relevant space permissions (read, administer, delete)
        """

        # Separate flags to track if view or edit is restricted at the page level
        page_view_restricted = bool(
            doc_permissions.get("read", {}).get("restrictions", {}).get("user", {}).get("results")
            or doc_permissions.get("read", {})
            .get("restrictions", {})
            .get("group", {})
            .get("results")
        )

        page_edit_restricted = bool(
            doc_permissions.get("update", {}).get("restrictions", {}).get("user", {}).get("results")
            or doc_permissions.get("update", {})
            .get("restrictions", {})
            .get("group", {})
            .get("results")
        )

        permissions_by_role = {
            "read": {"users": set(), "groups": set()},
            "update": {"users": set(), "groups": set()},
            "delete": {"users": set(), "groups": set()},
        }

        total_permissions = 0

        for action, permissions in doc_permissions.items():
            restrictions_dict = permissions.get("restrictions", {})

            for entity_type, entity_data in restrictions_dict.items():
                for entity in entity_data.get("results"):
                    entity_id = entity["accountId"] if entity_type == "user" else entity["id"]
                    permissions_by_role[action][f"{entity_type}s"].add(entity_id)
                    total_permissions += 1
                    # edit permission implies view permission
                    if action == "update":
                        permissions_by_role["read"][f"{entity_type}s"].add(entity_id)
                        # total_permissions += 1
                        # ^ omitting to not double count an entity.
                        # may result in a higher total count than max_num_metadata_permissions

        for space_perm in space_permissions:
            if total_permissions < self.download_config.max_num_metadata_permissions:
                space_operation = space_perm["operation"]["key"]
                space_target_type = space_perm["operation"]["targetType"]
                space_entity_id = space_perm["principal"]["id"]
                space_entity_type = space_perm["principal"]["type"]

                # Apply space-level view permissions if no page restrictions exist
                if (
                    space_target_type == "space"
                    and space_operation == "read"
                    and not page_view_restricted
                ):
                    permissions_by_role["read"][f"{space_entity_type}s"].add(space_entity_id)
                    total_permissions += 1

                # Administer permission includes view + edit. Apply if not page restricted
                elif space_target_type == "space" and space_operation == "administer":
                    if not page_view_restricted:
                        permissions_by_role["read"][f"{space_entity_type}s"].add(space_entity_id)
                        total_permissions += 1
                        if not page_edit_restricted:
                            permissions_by_role["update"][f"{space_entity_type}s"].add(
                                space_entity_id
                            )
                            # total_permissions += 1
                            # ^ omitting to not double count an entity.
                            # may result in a higher total count than max_num_metadata_permissions

                # Add the "delete page" space permissions if there are other page permissions
                elif (
                    space_target_type == "page"
                    and space_operation == "delete"
                    and space_entity_id in permissions_by_role["read"][f"{space_entity_type}s"]
                ):
                    permissions_by_role["delete"][f"{space_entity_type}s"].add(space_entity_id)
                    total_permissions += 1

        # turn sets into sorted lists for consistency and json serialization
        for role_dict in permissions_by_role.values():
            for key in role_dict:
                role_dict[key] = sorted(role_dict[key])

        return permissions_by_role

    def _get_permissions_for_space(self, space_id: int) -> Optional[List[dict]]:
        if space_id in self._permissions_cache:
            self._permissions_cache.move_to_end(space_id)  # mark recent use
            logger.debug(f"Retrieved cached permissions for space {space_id}")
            return self._permissions_cache[space_id]
        else:
            with self.connection_config.get_client() as client:
                try:
                    # TODO limit the total number of results being called.
                    # not yet implemented because this client call doesn't allow for filtering for
                    # certain operations, so adding a limit here would result in too little data.
                    space_permissions = []
                    space_permissions_result = client.get(f"/api/v2/spaces/{space_id}/permissions")
                    space_permissions.extend(space_permissions_result["results"])
                    if space_permissions_result["_links"].get("next"):  # pagination
                        while space_permissions_result.get("next"):
                            space_permissions_result = client.get(space_permissions_result["next"])
                            space_permissions.extend(space_permissions_result["results"])

                    if len(self._permissions_cache) >= self._permissions_cache_max_size:
                        self._permissions_cache.popitem(last=False)  # LRU/FIFO eviction
                    self._permissions_cache[space_id] = space_permissions

                    logger.debug(f"Retrieved permissions for space {space_id}")
                    return space_permissions
                except Exception as e:
                    logger.debug(f"Could not retrieve permissions for space {space_id}: {e}")
                    return None

    def _parse_permissions_for_doc(
        self, doc_id: str, space_permissions: list
    ) -> Optional[list[dict]]:
        with self.connection_config.get_client() as client:
            try:
                doc_permissions = client.get_all_restrictions_for_content(content_id=doc_id)
                parsed_permissions_dict = self.parse_permissions(doc_permissions, space_permissions)
                parsed_permissions_dict = [{k: v} for k, v in parsed_permissions_dict.items()]

            except Exception as e:
                # skip writing any permission metadata
                logger.debug(f"Could not retrieve permissions for doc {doc_id}: {e}")
                return None

        logger.debug(f"normalized permissions generated: {parsed_permissions_dict}")
        return parsed_permissions_dict

    def run(self, file_data: FileData, **kwargs) -> download_responses:
        from bs4 import BeautifulSoup

        doc_id = file_data.identifier
        try:
            with self.connection_config.get_client() as client:
                page = client.get_page_by_id(
                    page_id=doc_id,
                    expand="history.lastUpdated,version,body.view",
                )
        except Exception as e:
            logger.exception(f"Failed to retrieve page with ID {doc_id}: {e}")
            raise SourceConnectionError(f"Failed to retrieve page with ID {doc_id}: {e}")

        if not page:
            raise ValueError(f"Page with ID {doc_id} does not exist.")

        content = page["body"]["view"]["value"]
        title = page["title"]
        # Using h1 for title is supported by both v1 and v2 html parsing in unstructured
        title_html = f"<h1>{title}</h1>"
        content = fix_unescaped_unicode(f"<body class='Document' >{title_html}{content}</body>")
        if self.download_config.extract_images:
            with self.connection_config.get_client() as client:
                content = self.download_config.extract_html_images(
                    url=file_data.metadata.url, html=content, session=client._session
                )

        filepath = file_data.source_identifiers.relative_path
        download_path = Path(self.download_dir) / filepath
        download_path.parent.mkdir(parents=True, exist_ok=True)
        with open(download_path, "w", encoding="utf8") as f:
            soup = BeautifulSoup(content, "html.parser")
            f.write(soup.prettify())

        # Get document permissions and update metadata
        space_id = file_data.additional_metadata["space_id"]
        space_perm = self._get_permissions_for_space(space_id)  # must be the id, NOT the space key
        if space_perm:
            combined_doc_permissions = self._parse_permissions_for_doc(doc_id, space_perm)
            if combined_doc_permissions:
                file_data.metadata.permissions_data = combined_doc_permissions

        # Update file_data with metadata
        file_data.metadata.date_created = page["history"]["createdDate"]
        file_data.metadata.date_modified = page["version"]["when"]
        file_data.metadata.version = str(page["version"]["number"])
        file_data.display_name = title

        download_response = self.generate_download_response(
            file_data=file_data, download_path=download_path
        )
        if self.download_config.extract_files:
            with self.connection_config.get_client() as client:
                extracted_download_responses = self.download_embedded_files(
                    html=content,
                    current_file_data=download_response["file_data"],
                    session=client._session,
                )
                if extracted_download_responses:
                    for dr in extracted_download_responses:
                        fd = dr["file_data"]
                        source_file_path = Path(file_data.source_identifiers.fullpath).with_suffix(
                            ""
                        )
                        new_fullpath = source_file_path / fd.source_identifiers.filename
                        fd.source_identifiers = SourceIdentifiers(
                            fullpath=new_fullpath.as_posix(), filename=new_fullpath.name
                        )
                    extracted_download_responses.append(download_response)
                    return extracted_download_responses
        return download_response


confluence_source_entry = SourceRegistryEntry(
    connection_config=ConfluenceConnectionConfig,
    indexer_config=ConfluenceIndexerConfig,
    indexer=ConfluenceIndexer,
    downloader_config=ConfluenceDownloaderConfig,
    downloader=ConfluenceDownloader,
)



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional, Union

from dateutil import parser
from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileDataSourceMetadata
from unstructured_ingest.errors_v2 import ProviderError, UserError
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)
from unstructured_ingest.processes.utils.blob_storage import (
    BlobStoreUploadStager,
    BlobStoreUploadStagerConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies
from unstructured_ingest.utils.string_and_date_utils import json_to_dict

if TYPE_CHECKING:
    from gcsfs import GCSFileSystem

CONNECTOR_TYPE = "gcs"


class GcsIndexerConfig(FsspecIndexerConfig):
    pass


service_account_key_description = """
  Options:
  - ``None``, GCSFS will attempt to guess your credentials in the
following order: gcloud CLI default, gcsfs cached token, google compute
metadata service, anonymous.
  - ``'google_default'``, your default gcloud credentials will be used,
    which are typically established by doing ``gcloud login`` in a terminal.
  - ``'cache'``, credentials from previously successful gcsfs
    authentication will be used (use this after "browser" auth succeeded)
  - ``'anon'``, no authentication is performed, and you can only
    access data which is accessible to allUsers (in this case, the project and
    access level parameters are meaningless)
  - ``'browser'``, you get an access code with which you can
    authenticate via a specially provided URL
  - if ``'cloud'``, we assume we are running within google compute
    or google container engine, and query the internal metadata directly for
    a token.
  - you may supply a token generated by the
    [gcloud](https://cloud.google.com/sdk/docs/)
    utility; this is either a python dictionary or the name of a file
    containing the JSON returned by logging in with the gcloud CLI tool.
  """


class GcsAccessConfig(FsspecAccessConfig):
    service_account_key: Optional[str] = Field(
        default=None, description=service_account_key_description
    )
    token: Union[str, dict, None] = Field(init=False, default=None)

    def model_post_init(self, __context: Any) -> None:
        ALLOWED_AUTH_VALUES = "google_default", "cache", "anon", "browser", "cloud"

        # Case: null value
        if not self.service_account_key:
            return

        # Case: one of auth constants
        if self.service_account_key in ALLOWED_AUTH_VALUES:
            self.token = self.service_account_key
            return

        # Case: token as json
        if isinstance(json_to_dict(self.service_account_key), dict):
            self.token = json_to_dict(self.service_account_key)
            return

        # Case: path to token
        if Path(self.service_account_key).is_file():
            self.token = self.service_account_key
            return

        raise ValueError("Invalid auth token value")


class GcsConnectionConfig(FsspecConnectionConfig):
    supported_protocols: list[str] = field(default_factory=lambda: ["gs", "gcs"], init=False)
    access_config: Secret[GcsAccessConfig] = Field(default=GcsAccessConfig(), validate_default=True)
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    @requires_dependencies(["gcsfs", "fsspec"], extras="gcs")
    @contextmanager
    def get_client(self, protocol: str) -> Generator["GCSFileSystem", None, None]:
        with super().get_client(protocol=protocol) as client:
            yield client

    def wrap_error(self, e: Exception) -> Exception:
        # https://github.com/fsspec/gcsfs/blob/main/gcsfs/retry.py#L79
        from gcsfs.retry import HttpError

        if isinstance(e, FileNotFoundError):
            raise UserError(f"File not found: {e}")
        if isinstance(e, OSError) and "Forbidden" in str(e):
            raise UserError(e)
        if isinstance(e, ValueError) and "Bad Request" in str(e):
            raise UserError(e)
        if isinstance(e, HttpError) and (http_error_code := e.code):
            message = e.message or e
            if 400 <= http_error_code < 500:
                raise UserError(message)
            if http_error_code >= 500:
                raise ProviderError(message)
        logger.error(f"unhandled exception from gcs ({type(e)}): {e}", exc_info=True)
        return e


@dataclass
class GcsIndexer(FsspecIndexer):
    connection_config: GcsConnectionConfig
    index_config: GcsIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def get_metadata(self, file_info: dict) -> FileDataSourceMetadata:
        path = file_info["name"]
        date_created = None
        date_modified = None
        if modified_at_str := file_info.get("updated"):
            date_modified = str(parser.parse(modified_at_str).timestamp())
        if created_at_str := file_info.get("timeCreated"):
            date_created = str(parser.parse(created_at_str).timestamp())

        file_size = file_info.get("size") if "size" in file_info else None

        version = file_info.get("etag")
        record_locator = {
            "protocol": self.index_config.protocol,
            "remote_file_path": self.index_config.remote_url,
            "file_id": file_info.get("id"),
        }
        return FileDataSourceMetadata(
            date_created=date_created,
            date_modified=date_modified,
            date_processed=str(time()),
            version=version,
            url=f"{self.index_config.protocol}://{path}",
            record_locator=record_locator,
            filesize_bytes=file_size,
        )


class GcsDownloaderConfig(FsspecDownloaderConfig):
    pass


@dataclass
class GcsDownloader(FsspecDownloader):
    protocol: str = "gcs"
    connection_config: GcsConnectionConfig
    connector_type: str = CONNECTOR_TYPE
    download_config: Optional[GcsDownloaderConfig] = field(default_factory=GcsDownloaderConfig)


class GcsUploaderConfig(FsspecUploaderConfig):
    pass


@dataclass
class GcsUploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: GcsConnectionConfig
    upload_config: GcsUploaderConfig = field(default=None)


gcs_source_entry = SourceRegistryEntry(
    indexer=GcsIndexer,
    indexer_config=GcsIndexerConfig,
    downloader=GcsDownloader,
    downloader_config=GcsDownloaderConfig,
    connection_config=GcsConnectionConfig,
)

gcs_destination_entry = DestinationRegistryEntry(
    uploader=GcsUploader,
    uploader_config=GcsUploaderConfig,
    connection_config=GcsConnectionConfig,
    upload_stager_config=BlobStoreUploadStagerConfig,
    upload_stager=BlobStoreUploadStager,
)



import json
import os
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
)
from unstructured_ingest.processes.connectors.sql.sql import (
    SQLAccessConfig,
    SQLConnectionConfig,
    SQLUploader,
    SQLUploaderConfig,
    SQLUploadStager,
    SQLUploadStagerConfig,
)
from unstructured_ingest.utils.data_prep import split_dataframe
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from databricks.sdk.core import oauth_service_principal
    from databricks.sql.client import Connection as DeltaTableConnection
    from databricks.sql.client import Cursor as DeltaTableCursor
    from pandas import DataFrame

CONNECTOR_TYPE = "databricks_delta_tables"


class DatabricksDeltaTablesAccessConfig(SQLAccessConfig):
    token: Optional[str] = Field(default=None, description="Databricks Personal Access Token")
    client_id: Optional[str] = Field(default=None, description="Client ID of the OAuth app.")
    client_secret: Optional[str] = Field(
        default=None, description="Client Secret of the OAuth app."
    )


class DatabricksDeltaTablesConnectionConfig(SQLConnectionConfig):
    access_config: Secret[DatabricksDeltaTablesAccessConfig]
    server_hostname: str = Field(description="server hostname connection config value")
    http_path: str = Field(description="http path connection config value")

    @requires_dependencies(["databricks"], extras="databricks-delta-tables")
    def get_credentials_provider(self) -> "oauth_service_principal":
        from databricks.sdk.core import Config, oauth_service_principal

        host = f"https://{self.server_hostname}"
        access_configs = self.access_config.get_secret_value()
        client_id = access_configs.client_id
        client_secret = access_configs.client_secret

        def _get_credentials_provider():
            return oauth_service_principal(
                Config(
                    host=host,
                    client_id=client_id,
                    client_secret=client_secret,
                )
            )

        if client_id and client_secret:
            return _get_credentials_provider

        return False

    def model_post_init(self, __context: Any) -> None:
        access_config = self.access_config.get_secret_value()
        if access_config.token and access_config.client_secret and access_config.client_id:
            raise ValueError(
                "One one for of auth can be provided, either token or client id and secret"
            )
        if not access_config.token and not (
            access_config.client_secret and access_config.client_id
        ):
            raise ValueError(
                "One form of auth must be provided, either token or client id and secret"
            )

    @contextmanager
    @requires_dependencies(["databricks"], extras="databricks-delta-tables")
    def get_connection(self, **connect_kwargs) -> Generator["DeltaTableConnection", None, None]:
        from databricks.sql import connect

        connect_kwargs = connect_kwargs or {}
        connect_kwargs["_user_agent_entry"] = os.getenv(
            "UNSTRUCTURED_USER_AGENT", "unstructuredio_oss"
        )
        connect_kwargs["server_hostname"] = connect_kwargs.get(
            "server_hostname", self.server_hostname
        )
        connect_kwargs["http_path"] = connect_kwargs.get("http_path", self.http_path)

        if credential_provider := self.get_credentials_provider():
            connect_kwargs["credentials_provider"] = credential_provider
        else:
            connect_kwargs["access_token"] = self.access_config.get_secret_value().token
        with connect(**connect_kwargs) as connection:
            yield connection

    @contextmanager
    def get_cursor(self, **connect_kwargs) -> Generator["DeltaTableCursor", None, None]:
        with self.get_connection(**connect_kwargs) as connection:
            cursor = connection.cursor()
            yield cursor


class DatabricksDeltaTablesUploadStagerConfig(SQLUploadStagerConfig):
    pass


class DatabricksDeltaTablesUploadStager(SQLUploadStager):
    upload_stager_config: DatabricksDeltaTablesUploadStagerConfig


class DatabricksDeltaTablesUploaderConfig(SQLUploaderConfig):
    catalog: str = Field(description="Name of the catalog in the Databricks Unity Catalog service")
    database: str = Field(description="Database name", default="default")
    table_name: str = Field(description="Table name")


@dataclass
class DatabricksDeltaTablesUploader(SQLUploader):
    upload_config: DatabricksDeltaTablesUploaderConfig
    connection_config: DatabricksDeltaTablesConnectionConfig
    connector_type: str = CONNECTOR_TYPE

    @requires_dependencies(["pandas"], extras="databricks-delta-tables")
    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        super().run(path=path, file_data=file_data, **kwargs)

    @contextmanager
    def get_cursor(self) -> Generator[Any, None, None]:
        with self.connection_config.get_cursor() as cursor:
            cursor.execute(f"USE CATALOG '{self.upload_config.catalog}'")
            yield cursor

    def precheck(self) -> None:
        with self.connection_config.get_cursor() as cursor:
            cursor.execute("SHOW CATALOGS")
            catalogs = [r[0] for r in cursor.fetchall()]
            if self.upload_config.catalog not in catalogs:
                raise ValueError(
                    "Catalog {} not found in {}".format(
                        self.upload_config.catalog, ", ".join(catalogs)
                    )
                )
            cursor.execute(f"USE CATALOG '{self.upload_config.catalog}'")
            cursor.execute("SHOW DATABASES")
            databases = [r[0] for r in cursor.fetchall()]
            if self.upload_config.database not in databases:
                raise ValueError(
                    "Database {} not found in {}".format(
                        self.upload_config.database, ", ".join(databases)
                    )
                )
            cursor.execute("SHOW TABLES")
            table_names = [r[1] for r in cursor.fetchall()]
            if self.upload_config.table_name not in table_names:
                raise ValueError(
                    "Table {} not found in {}".format(
                        self.upload_config.table_name, ", ".join(table_names)
                    )
                )

    def create_statement(self, columns: list[str], values: tuple[Any, ...]) -> str:
        values_list = []
        for v in values:
            if isinstance(v, dict):
                values_list.append(json.dumps(v))
            elif isinstance(v, list):
                if v and isinstance(v[0], (int, float)):
                    values_list.append("ARRAY({})".format(", ".join([str(val) for val in v])))
                else:
                    values_list.append("ARRAY({})".format(", ".join([f"'{val}'" for val in v])))
            else:
                values_list.append(f"'{v}'")
        statement = "INSERT INTO {table_name} ({columns}) VALUES({values})".format(
            table_name=self.upload_config.table_name,
            columns=", ".join(columns),
            values=", ".join(values_list),
        )
        return statement

    @requires_dependencies(["pandas"], extras="databricks-delta-tables")
    def upload_dataframe(self, df: "DataFrame", file_data: FileData) -> None:
        import numpy as np

        if self.can_delete():
            self.delete_by_record_id(file_data=file_data)
        else:
            logger.warning(
                f"table doesn't contain expected "
                f"record id column "
                f"{self.upload_config.record_id_key}, skipping delete"
            )
        df.replace({np.nan: None}, inplace=True)
        self._fit_to_schema(df=df)

        columns = list(df.columns)
        logger.info(
            f"writing a total of {len(df)} elements via"
            f" document batches to destination"
            f" table named {self.upload_config.table_name}"
            # f" with batch size {self.upload_config.batch_size}"
        )
        # TODO: currently variable binding not supporting for list data_types,
        #  update once that gets resolved in SDK
        for rows in split_dataframe(df=df, chunk_size=self.upload_config.batch_size):
            with self.get_cursor() as cursor:
                values = self.prepare_data(columns, tuple(rows.itertuples(index=False, name=None)))
                for v in values:
                    stmt = self.create_statement(columns=columns, values=v)
                    cursor.execute(stmt)


databricks_delta_tables_destination_entry = DestinationRegistryEntry(
    connection_config=DatabricksDeltaTablesConnectionConfig,
    uploader=DatabricksDeltaTablesUploader,
    uploader_config=DatabricksDeltaTablesUploaderConfig,
    upload_stager=DatabricksDeltaTablesUploadStager,
    upload_stager_config=DatabricksDeltaTablesUploadStagerConfig,
)



# https://developers.notion.com/reference/property-object#title
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import DBCellBase, DBPropertyBase
from unstructured_ingest.processes.connectors.notion.types.rich_text import RichText


@dataclass
class Title(DBPropertyBase):
    id: str
    name: str
    type: str = "title"
    title: dict = field(default_factory=dict)
    description: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class TitleCell(DBCellBase):
    id: str
    title: List[RichText]
    type: str = "title"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(title=[RichText.from_dict(rt) for rt in data.pop("title", [])], **data)

    def get_html(self) -> Optional[HtmlTag]:
        if not self.title:
            return None
        return Div([], [rt.get_html() for rt in self.title])



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, AsyncIterator, Optional

from pydantic import Field

from unstructured_ingest.data_types.file_data import (
    FileData,
)
from unstructured_ingest.error import (
    SourceConnectionError,
    SourceConnectionNetworkError,
)
from unstructured_ingest.errors_v2 import UserAuthError, UserError
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    SourceRegistryEntry,
)
from unstructured_ingest.processes.connectors.onedrive import (
    OnedriveAccessConfig,
    OnedriveConnectionConfig,
    OnedriveDownloader,
    OnedriveDownloaderConfig,
    OnedriveIndexer,
    OnedriveIndexerConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from office365.onedrive.driveitems.driveItem import DriveItem
    from office365.onedrive.sites.site import Site
    from office365.runtime.client_request_exception import ClientRequestException

CONNECTOR_TYPE = "sharepoint"
LEGACY_DEFAULT_PATH = "Shared Documents"


class SharepointAccessConfig(OnedriveAccessConfig):
    client_cred: str = Field(description="Microsoft App client secret")


class SharepointConnectionConfig(OnedriveConnectionConfig):
    user_pname: Optional[str] = Field(
        default=None,
        description="User principal name or service account, usually your Azure AD email.",
    )
    site: str = Field(
        description="Sharepoint site url. Process either base url e.g \
                    https://[tenant].sharepoint.com  or relative sites \
                    https://[tenant].sharepoint.com/sites/<site_name>. \
                    To process all sites within the tenant pass a site url as \
                    https://[tenant]-admin.sharepoint.com.\
                    This requires the app to be registered at a tenant level"
    )
    library: Optional[str] = Field(
        default=None,
        description="Sharepoint library name. If not provided, the default \
                    drive will be used.",
    )

    def _get_drive_item(self, client_site: Site) -> DriveItem:
        """Helper method to get the drive item for the specified library or default drive."""
        site_drive_item = None
        if self.library:
            for drive in client_site.drives.get().execute_query():
                if drive.name == self.library:
                    logger.info(f"Found the requested library: {self.library}")
                    site_drive_item = drive.get().execute_query().root
                    break

        # If no specific library was found or requested, use the default drive
        if not site_drive_item:
            if self.library:
                logger.warning(
                    f"Library '{self.library}' not found in site '{self.site}'. "
                    "Using the default drive instead."
                )

            site_drive_item = client_site.drive.get().execute_query().root

        return site_drive_item


class SharepointIndexerConfig(OnedriveIndexerConfig):
    # TODO: We can probably make path non-optional on OnedriveIndexerConfig once tested
    path: str = Field(default="")

@dataclass
class SharepointIndexer(OnedriveIndexer):
    connection_config: SharepointConnectionConfig
    index_config: SharepointIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def _handle_client_request_exception(self, e: ClientRequestException, context: str) -> None:
        """Convert ClientRequestException to appropriate user-facing error based on HTTP status."""
        if hasattr(e, "response") and e.response is not None and hasattr(e.response, "status_code"):
            status_code = e.response.status_code
            if status_code == 401:
                raise UserAuthError(
                    f"Unauthorized access to {context}. Check client credentials and permissions"
                )
            elif status_code == 403:
                raise UserAuthError(
                    f"Access forbidden to {context}. "
                    f"Check app permissions (Sites.Read.All required)"
                )
            elif status_code == 404:
                raise UserError(f"Not found: {context}")

        raise UserError(f"Failed to access {context}: {str(e)}")

    def _is_root_path(self, path: str) -> bool:
        """Check if the path represents root access (empty string or legacy default)."""
        return not path or not path.strip() or path == LEGACY_DEFAULT_PATH
    
    def _get_target_drive_item(self, site_drive_item: DriveItem, path: str) -> DriveItem:
        """Get the drive item to search in based on the path."""
        if self._is_root_path(path):
            return site_drive_item
        else:
            return site_drive_item.get_by_path(path).get().execute_query()
    
    def _validate_folder_path(self, site_drive_item: DriveItem, path: str) -> None:
        """Validate that a specific folder path exists and is accessible."""
        from office365.runtime.client_request_exception import ClientRequestException

        try:
            path_item = site_drive_item.get_by_path(path).get().execute_query()
            if path_item is None or not hasattr(path_item, "is_folder"):
                raise UserError(
                    f"SharePoint path '{path}' not found in site {self.connection_config.site}. "
                    f"Check that the path exists and you have access to it"
                )
            logger.info(f"SharePoint folder path '{path}' validated successfully")
        except ClientRequestException as e:
            logger.error(f"Failed to access SharePoint path '{path}': {e}")
            self._handle_client_request_exception(e, f"SharePoint path '{path}'")
        except Exception as e:
            logger.error(f"Unexpected error accessing SharePoint path '{path}': {e}")
            raise UserError(f"Failed to validate SharePoint path '{path}': {str(e)}")

    @requires_dependencies(["office365"], extras="sharepoint")
    def precheck(self) -> None:
        """Validate SharePoint connection before indexing."""
        from office365.runtime.client_request_exception import ClientRequestException

        # Validate authentication - this call will raise UserAuthError if invalid
        self.connection_config.get_token()

        try:
            client = self.connection_config.get_client()
            client_site = client.sites.get_by_url(self.connection_config.site).get().execute_query()
            site_drive_item = self.connection_config._get_drive_item(client_site)

            path = self.index_config.path
            if not self._is_root_path(path):
                self._validate_folder_path(site_drive_item, path)

            logger.info(
                f"SharePoint connection validated successfully for site: "
                f"{self.connection_config.site}"
            )

        except ClientRequestException as e:
            logger.error(f"SharePoint precheck failed for site: {self.connection_config.site}")
            self._handle_client_request_exception(
                e, f"SharePoint site {self.connection_config.site}"
            )
        except Exception as e:
            logger.error(f"Unexpected error during SharePoint precheck: {e}", exc_info=True)
            raise UserError(f"Failed to validate SharePoint connection: {str(e)}")

    @requires_dependencies(["office365"], extras="sharepoint")
    async def run_async(self, **kwargs: Any) -> AsyncIterator[FileData]:
        from office365.runtime.client_request_exception import ClientRequestException

        token_resp = await asyncio.to_thread(self.connection_config.get_token)
        if "error" in token_resp:
            raise SourceConnectionError(
                f"[{self.connector_type}]: {token_resp['error']} "
                f"({token_resp.get('error_description')})"
            )

        client = await asyncio.to_thread(self.connection_config.get_client)
        try:
            client_site = client.sites.get_by_url(self.connection_config.site).get().execute_query()
            site_drive_item = self.connection_config._get_drive_item(client_site)
        except ClientRequestException as e:
            logger.error(f"Failed to access SharePoint site: {self.connection_config.site}")
            raise SourceConnectionError(
                f"Unable to access SharePoint site at {self.connection_config.site}: {str(e)}"
            )

        path = self.index_config.path
        target_drive_item = await asyncio.to_thread(
            self._get_target_drive_item, site_drive_item, path
        )

        for drive_item in target_drive_item.get_files(
            recursive=self.index_config.recursive
        ).execute_query():
            file_data = await self.drive_item_to_file_data(drive_item=drive_item)
            yield file_data


class SharepointDownloaderConfig(OnedriveDownloaderConfig):
    pass


@dataclass
class SharepointDownloader(OnedriveDownloader):
    connection_config: SharepointConnectionConfig
    download_config: SharepointDownloaderConfig
    connector_type: str = CONNECTOR_TYPE

    @SourceConnectionNetworkError.wrap
    @requires_dependencies(["office365"], extras="sharepoint")
    def _fetch_file(self, file_data: FileData) -> DriveItem:
        from office365.runtime.client_request_exception import ClientRequestException

        if file_data.source_identifiers is None or not file_data.source_identifiers.fullpath:
            raise ValueError(
                f"file data doesn't have enough information to get "
                f"file content: {file_data.model_dump()}"
            )

        server_relative_path = file_data.source_identifiers.fullpath
        client = self.connection_config.get_client()

        try:
            client_site = client.sites.get_by_url(self.connection_config.site).get().execute_query()
            site_drive_item = self.connection_config._get_drive_item(client_site)
        except ClientRequestException:
            logger.info("Site not found")
            raise SourceConnectionError(f"Site not found: {self.connection_config.site}")
        file = site_drive_item.get_by_path(server_relative_path).get().execute_query()

        if not file:
            raise FileNotFoundError(f"file not found: {server_relative_path}")
        return file


sharepoint_source_entry = SourceRegistryEntry(
    connection_config=SharepointConnectionConfig,
    indexer_config=SharepointIndexerConfig,
    indexer=SharepointIndexer,
    downloader_config=SharepointDownloaderConfig,
    downloader=SharepointDownloader,
)



from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.weaviate.weaviate import (
    WeaviateAccessConfig,
    WeaviateConnectionConfig,
    WeaviateUploader,
    WeaviateUploaderConfig,
    WeaviateUploadStager,
    WeaviateUploadStagerConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from weaviate.client import WeaviateClient

CONNECTOR_TYPE = "weaviate-embedded"


class EmbeddedWeaviateAccessConfig(WeaviateAccessConfig):
    pass


class EmbeddedWeaviateConnectionConfig(WeaviateConnectionConfig):
    hostname: str = Field(default="127.0.0.1", description="hostname")
    port: int = Field(default=8079, description="http port")
    grpc_port: int = Field(default=50050, description="grpc port")
    data_path: Optional[str] = Field(
        default=None,
        description="directory where the files making up the "
        "database are stored. If not provided, will "
        "default to underlying SDK implementation",
    )
    access_config: Secret[WeaviateAccessConfig] = Field(
        default=WeaviateAccessConfig(), validate_default=True
    )

    @contextmanager
    @requires_dependencies(["weaviate"], extras="weaviate")
    def get_client(self) -> Generator["WeaviateClient", None, None]:
        from weaviate import connect_to_embedded
        from weaviate.classes.init import AdditionalConfig

        with connect_to_embedded(
            hostname=self.hostname,
            port=self.port,
            grpc_port=self.grpc_port,
            persistence_data_path=self.data_path,
            additional_config=AdditionalConfig(timeout=self.get_timeout()),
        ) as weaviate_client:
            yield weaviate_client


class EmbeddedWeaviateUploadStagerConfig(WeaviateUploadStagerConfig):
    pass


@dataclass
class EmbeddedWeaviateUploadStager(WeaviateUploadStager):
    upload_stager_config: EmbeddedWeaviateUploadStagerConfig = field(
        default_factory=lambda: WeaviateUploadStagerConfig()
    )


class EmbeddedWeaviateUploaderConfig(WeaviateUploaderConfig):
    pass


@dataclass
class EmbeddedWeaviateUploader(WeaviateUploader):
    connection_config: EmbeddedWeaviateConnectionConfig = field(
        default_factory=lambda: EmbeddedWeaviateConnectionConfig()
    )
    upload_config: EmbeddedWeaviateUploaderConfig = field(
        default_factory=lambda: EmbeddedWeaviateUploaderConfig()
    )
    connector_type: str = CONNECTOR_TYPE


weaviate_embedded_destination_entry = DestinationRegistryEntry(
    connection_config=EmbeddedWeaviateConnectionConfig,
    uploader=EmbeddedWeaviateUploader,
    uploader_config=EmbeddedWeaviateUploaderConfig,
    upload_stager=EmbeddedWeaviateUploadStager,
    upload_stager_config=EmbeddedWeaviateUploadStagerConfig,
)



import logging
from dataclasses import dataclass, field
from typing import Any

import click
from pydantic import BaseModel

from unstructured_ingest.cli.base.cmd import BaseCmd
from unstructured_ingest.cli.utils.click import Group, conform_click_options
from unstructured_ingest.cli.utils.model_conversion import options_from_base_model
from unstructured_ingest.interfaces import ProcessorConfig
from unstructured_ingest.logger import logger
from unstructured_ingest.processes import (
    ChunkerConfig,
    EmbedderConfig,
    FiltererConfig,
    PartitionerConfig,
)
from unstructured_ingest.processes.connector_registry import SourceRegistryEntry


@dataclass
class SrcCmd(BaseCmd):
    registry_entry: SourceRegistryEntry
    default_configs: list[BaseModel] = field(
        default_factory=lambda: [
            ProcessorConfig,
            PartitionerConfig,
            EmbedderConfig,
            FiltererConfig,
            ChunkerConfig,
        ]
    )

    def get_registry_options(self):
        options = []
        configs = [
            config
            for config in [
                self.registry_entry.connection_config,
                self.registry_entry.indexer_config,
                self.registry_entry.downloader_config,
            ]
            if config
        ]
        for config in configs:
            options.extend(options_from_base_model(model=config))
        options = self.consolidate_options(options=options)
        return options

    def cmd(self, ctx: click.Context, **options: dict[str, Any]) -> None:
        if ctx.invoked_subcommand:
            return

        conform_click_options(options)
        logger.setLevel(logging.DEBUG if options.get("verbose", False) else logging.INFO)
        try:
            pipeline = self.get_pipeline(src=self.cmd_name, source_options=options)
            pipeline.run()
        except Exception as e:
            logger.error(f"failed to run source command {self.cmd_name}: {e}", exc_info=True)
            raise click.ClickException(str(e)) from e

    def get_cmd(self) -> click.Group:
        # Dynamically create the command without the use of click decorators
        fn = self.cmd
        fn = click.pass_context(fn)
        cmd = click.group(fn, cls=Group)
        if not isinstance(cmd, click.core.Group):
            raise ValueError(f"generated src command was not of expected type Group: {type(cmd)}")
        cmd.name = self.cli_cmd_name
        cmd.invoke_without_command = True
        self.add_options(cmd)

        return cmd



import hashlib
import json
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, Any, Generator, Union

from dateutil import parser
from pydantic import BaseModel, Field, Secret

from unstructured_ingest.data_types.file_data import (
    BatchFileData,
    BatchItem,
    FileData,
    FileDataSourceMetadata,
    SourceIdentifiers,
)
from unstructured_ingest.error import DestinationConnectionError, SourceConnectionError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Downloader,
    DownloaderConfig,
    DownloadResponse,
    Indexer,
    IndexerConfig,
    Uploader,
    UploaderConfig,
    UploadStager,
    UploadStagerConfig,
    download_responses,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.constants import RECORD_ID_LABEL
from unstructured_ingest.utils.data_prep import (
    get_data_df,
    get_enhanced_element_id,
    get_json_data,
    split_dataframe,
    write_data,
)

if TYPE_CHECKING:
    from pandas import DataFrame

_DATE_COLUMNS = ("date_created", "date_modified", "date_processed", "last_modified")


class SqlAdditionalMetadata(BaseModel):
    table_name: str
    id_column: str


class SqlBatchFileData(BatchFileData):
    additional_metadata: SqlAdditionalMetadata


def parse_date_string(date_value: Union[str, int]) -> datetime:
    try:
        timestamp = float(date_value) / 1000 if isinstance(date_value, int) else float(date_value)
        return datetime.fromtimestamp(timestamp)
    except Exception as e:
        logger.debug(f"date {date_value} string not a timestamp: {e}")
    return parser.parse(date_value)


class SQLAccessConfig(AccessConfig):
    pass


class SQLConnectionConfig(ConnectionConfig, ABC):
    access_config: Secret[SQLAccessConfig] = Field(default=SQLAccessConfig(), validate_default=True)

    @abstractmethod
    @contextmanager
    def get_connection(self) -> Generator[Any, None, None]:
        pass

    @abstractmethod
    @contextmanager
    def get_cursor(self) -> Generator[Any, None, None]:
        pass


class SQLIndexerConfig(IndexerConfig):
    table_name: str
    id_column: str
    batch_size: int = 100


class SQLIndexer(Indexer, ABC):
    connection_config: SQLConnectionConfig
    index_config: SQLIndexerConfig

    @contextmanager
    def get_cursor(self) -> Generator[Any, None, None]:
        with self.connection_config.get_cursor() as cursor:
            yield cursor

    def _get_doc_ids(self) -> list[str]:
        with self.get_cursor() as cursor:
            cursor.execute(
                f"SELECT {self.index_config.id_column} FROM {self.index_config.table_name}"
            )
            results = cursor.fetchall()
            ids = sorted([result[0] for result in results])
            return ids

    def precheck(self) -> None:
        try:
            with self.get_cursor() as cursor:
                cursor.execute("SELECT 1;")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise SourceConnectionError(f"failed to validate connection: {e}")

    def run(self, **kwargs: Any) -> Generator[SqlBatchFileData, None, None]:
        ids = self._get_doc_ids()
        id_batches: list[frozenset[str]] = [
            frozenset(
                ids[
                    i * self.index_config.batch_size : (i + 1)  # noqa
                    * self.index_config.batch_size
                ]
            )
            for i in range(
                (len(ids) + self.index_config.batch_size - 1) // self.index_config.batch_size
            )
        ]

        for batch in id_batches:
            batch_items = [BatchItem(identifier=str(b)) for b in batch]
            display_name = (
                f"{self.index_config.table_name}-{self.index_config.id_column}"
                f"-[{batch_items[0].identifier}..{batch_items[-1].identifier}]"
            )
            # Make sure the hash is always a positive number to create identified
            yield SqlBatchFileData(
                connector_type=self.connector_type,
                metadata=FileDataSourceMetadata(
                    date_processed=str(time()),
                ),
                additional_metadata=SqlAdditionalMetadata(
                    table_name=self.index_config.table_name, id_column=self.index_config.id_column
                ),
                batch_items=batch_items,
                display_name=display_name,
            )


class SQLDownloaderConfig(DownloaderConfig):
    fields: list[str] = field(default_factory=list)


class SQLDownloader(Downloader, ABC):
    connection_config: SQLConnectionConfig
    download_config: SQLDownloaderConfig

    @contextmanager
    def get_cursor(self) -> Generator[Any, None, None]:
        with self.connection_config.get_cursor() as cursor:
            yield cursor

    @abstractmethod
    def query_db(self, file_data: SqlBatchFileData) -> tuple[list[tuple], list[str]]:
        pass

    def sql_to_df(self, rows: list[tuple], columns: list[str]) -> list["DataFrame"]:
        import pandas as pd

        data = [dict(zip(columns, row)) for row in rows]
        df = pd.DataFrame(data)
        dfs = [pd.DataFrame([row.values], columns=df.columns) for index, row in df.iterrows()]
        return dfs

    def get_data(self, file_data: SqlBatchFileData) -> list["DataFrame"]:
        rows, columns = self.query_db(file_data=file_data)
        return self.sql_to_df(rows=rows, columns=columns)

    def get_identifier(self, table_name: str, record_id: str) -> str:
        f = f"{table_name}-{record_id}"
        if self.download_config.fields:
            f = "{}-{}".format(
                f,
                hashlib.sha256(",".join(self.download_config.fields).encode()).hexdigest()[:8],
            )
        return f

    def generate_download_response(
        self, result: "DataFrame", file_data: SqlBatchFileData
    ) -> DownloadResponse:
        id_column = file_data.additional_metadata.id_column
        table_name = file_data.additional_metadata.table_name
        record_id = result.iloc[0][id_column]
        filename_id = self.get_identifier(table_name=table_name, record_id=record_id)
        filename = f"{filename_id}.csv"
        download_path = self.download_dir / Path(filename)
        logger.debug(
            f"Downloading results from table {table_name} and id {record_id} to {download_path}"
        )
        download_path.parent.mkdir(parents=True, exist_ok=True)
        result.to_csv(download_path, index=False)
        file_data.source_identifiers = SourceIdentifiers(
            filename=filename,
            fullpath=filename,
        )
        cast_file_data = FileData.cast(file_data=file_data)
        cast_file_data.identifier = filename_id
        return super().generate_download_response(
            file_data=cast_file_data, download_path=download_path
        )

    def run(self, file_data: FileData, **kwargs: Any) -> download_responses:
        sql_filedata = SqlBatchFileData.cast(file_data=file_data)
        data_dfs = self.get_data(file_data=sql_filedata)
        download_responses = []
        for df in data_dfs:
            download_responses.append(
                self.generate_download_response(result=df, file_data=sql_filedata)
            )
        return download_responses


class SQLUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class SQLUploadStager(UploadStager):
    upload_stager_config: SQLUploadStagerConfig = field(default_factory=SQLUploadStagerConfig)

    def conform_dict(self, element_dict: dict, file_data: FileData) -> dict:
        data = element_dict.copy()
        metadata: dict[str, Any] = data.pop("metadata", {})
        data_source = metadata.pop("data_source", {})
        coordinates = metadata.pop("coordinates", {})

        data.update(metadata)
        data.update(data_source)
        data.update(coordinates)

        data["id"] = get_enhanced_element_id(element_dict=data, file_data=file_data)

        data[RECORD_ID_LABEL] = file_data.identifier
        return data

    def conform_dataframe(self, df: "DataFrame") -> "DataFrame":
        for column in filter(lambda x: x in df.columns, _DATE_COLUMNS):
            df[column] = df[column].apply(parse_date_string).apply(lambda date: date.timestamp())
        for column in filter(
            lambda x: x in df.columns,
            ("permissions_data", "record_locator", "points", "links"),
        ):
            df[column] = df[column].apply(
                lambda x: json.dumps(x) if isinstance(x, (list, dict)) else None
            )
        for column in filter(
            lambda x: x in df.columns,
            ("version", "page_number", "regex_metadata"),
        ):
            df[column] = df[column].apply(str)
        return df

    def write_output(self, output_path: Path, data: list[dict]) -> Path:
        write_data(path=output_path, data=data)
        return output_path

    def run(
        self,
        elements_filepath: Path,
        file_data: FileData,
        output_dir: Path,
        output_filename: str,
        **kwargs: Any,
    ) -> Path:
        import pandas as pd

        elements_contents = get_json_data(path=elements_filepath)

        df = pd.DataFrame(
            data=[
                self.conform_dict(element_dict=element_dict, file_data=file_data)
                for element_dict in elements_contents
            ]
        )
        df = self.conform_dataframe(df=df)

        output_filename_suffix = Path(elements_filepath).suffix
        output_filename = f"{Path(output_filename).stem}{output_filename_suffix}"
        output_path = self.get_output_path(output_filename=output_filename, output_dir=output_dir)

        final_output_path = self.write_output(
            output_path=output_path, data=df.to_dict(orient="records")
        )
        return final_output_path


class SQLUploaderConfig(UploaderConfig):
    batch_size: int = Field(default=50, description="Number of records per batch")
    table_name: str = Field(default="elements", description="which table to upload contents to")
    record_id_key: str = Field(
        default=RECORD_ID_LABEL,
        description="searchable key to find entries for the same record on previous runs",
    )


@dataclass
class SQLUploader(Uploader):
    upload_config: SQLUploaderConfig
    connection_config: SQLConnectionConfig
    values_delimiter: str = "?"
    _columns: list[str] = field(init=False, default=None)

    def precheck(self) -> None:
        try:
            with self.get_cursor() as cursor:
                cursor.execute("SELECT 1;")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    @contextmanager
    def get_cursor(self) -> Generator[Any, None, None]:
        with self.connection_config.get_cursor() as cursor:
            yield cursor

    def prepare_data(
        self, columns: list[str], data: tuple[tuple[Any, ...], ...]
    ) -> list[tuple[Any, ...]]:
        import pandas as pd

        output = []
        for row in data:
            parsed = []
            for column_name, value in zip(columns, row):
                if column_name in _DATE_COLUMNS:
                    if value is None or pd.isna(value):  # pandas is nan
                        parsed.append(None)
                    else:
                        parsed.append(parse_date_string(value))
                else:
                    parsed.append(value)
            output.append(tuple(parsed))
        return output

    def _fit_to_schema(
        self, df: "DataFrame", add_missing_columns: bool = True, case_sensitive: bool = True
    ) -> "DataFrame":
        import pandas as pd

        table_columns = self.get_table_columns()
        columns = set(df.columns if case_sensitive else df.columns.str.lower())
        schema_fields = set(
            table_columns if case_sensitive else {col.lower() for col in table_columns}
        )
        columns_to_drop = columns - schema_fields
        missing_columns = schema_fields - columns

        if columns_to_drop:
            logger.info(
                "Following columns will be dropped to match the table's schema: "
                f"{', '.join(columns_to_drop)}"
            )
        if missing_columns and add_missing_columns:
            logger.info(
                "Following null filled columns will be added to match the table's schema:"
                f" {', '.join(missing_columns)} "
            )

        df = df.drop(columns=columns_to_drop)

        if add_missing_columns:
            for column in missing_columns:
                df[column] = pd.Series()
        return df

    def upload_dataframe(self, df: "DataFrame", file_data: FileData) -> None:
        import numpy as np

        if self.can_delete():
            self.delete_by_record_id(file_data=file_data)
        else:
            logger.warning(
                f"table doesn't contain expected "
                f"record id column "
                f"{self.upload_config.record_id_key}, skipping delete"
            )
        df = self._fit_to_schema(df=df)
        df.replace({np.nan: None}, inplace=True)

        columns = list(df.columns)
        stmt = "INSERT INTO {table_name} ({columns}) VALUES({values})".format(
            table_name=self.upload_config.table_name,
            columns=",".join(columns),
            values=",".join([self.values_delimiter for _ in columns]),
        )
        logger.info(
            f"writing a total of {len(df)} elements via"
            f" document batches to destination"
            f" table named {self.upload_config.table_name}"
            f" with batch size {self.upload_config.batch_size}"
        )
        for rows in split_dataframe(df=df, chunk_size=self.upload_config.batch_size):
            with self.get_cursor() as cursor:
                values = self.prepare_data(columns, tuple(rows.itertuples(index=False, name=None)))
                # For debugging purposes:
                # for val in values:
                #     try:
                #         cursor.execute(stmt, val)
                #     except Exception as e:
                #         print(f"Error: {e}")
                #         print(f"failed to write {len(columns)}, {len(val)}: {stmt} -> {val}")
                logger.debug(f"running query: {stmt}")
                cursor.executemany(stmt, values)

    def get_table_columns(self) -> list[str]:
        if self._columns is None:
            with self.get_cursor() as cursor:
                cursor.execute(f"SELECT * from {self.upload_config.table_name} LIMIT 1")
                self._columns = [desc[0] for desc in cursor.description]
        return self._columns

    def can_delete(self) -> bool:
        return self.upload_config.record_id_key in self.get_table_columns()

    def delete_by_record_id(self, file_data: FileData) -> None:
        logger.debug(
            f"deleting any content with data "
            f"{self.upload_config.record_id_key}={file_data.identifier} "
            f"from table {self.upload_config.table_name}"
        )
        stmt = f"DELETE FROM {self.upload_config.table_name} WHERE {self.upload_config.record_id_key} = {self.values_delimiter}"  # noqa: E501
        with self.get_cursor() as cursor:
            cursor.execute(stmt, [file_data.identifier])
            rowcount = cursor.rowcount
            if rowcount > 0:
                logger.info(f"deleted {rowcount} rows from table {self.upload_config.table_name}")

    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        import pandas as pd

        df = pd.DataFrame(data)
        self.upload_dataframe(df=df, file_data=file_data)

    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        df = get_data_df(path=path)
        self.upload_dataframe(df=df, file_data=file_data)



# https://developers.notion.com/reference/page
from dataclasses import dataclass
from typing import Optional

from unstructured_ingest.processes.connectors.notion.interfaces import FromJSONMixin
from unstructured_ingest.processes.connectors.notion.types.file import FileObject
from unstructured_ingest.processes.connectors.notion.types.parent import Parent
from unstructured_ingest.processes.connectors.notion.types.user import PartialUser


@dataclass
class Page(FromJSONMixin):
    id: str
    created_time: str
    created_by: PartialUser
    last_edited_time: str
    last_edited_by: PartialUser
    archived: bool
    in_trash: bool
    properties: dict
    parent: Parent
    url: str
    public_url: str
    request_id: Optional[str] = None
    object: str = "page"
    icon: Optional[FileObject] = None
    cover: Optional[FileObject] = None

    @classmethod
    def from_dict(cls, data: dict):
        created_by = data.pop("created_by")
        last_edited_by = data.pop("last_edited_by")
        icon = data.pop("icon")
        cover = data.pop("cover")
        parent = data.pop("parent")
        page = cls(
            created_by=PartialUser.from_dict(created_by),
            last_edited_by=PartialUser.from_dict(last_edited_by),
            icon=FileObject.from_dict(icon) if icon else None,
            cover=FileObject.from_dict(cover) if cover else None,
            parent=Parent.from_dict(parent),
            **data,
        )

        return page



from dataclasses import dataclass

from pydantic import Field, Secret

from unstructured_ingest.interfaces.connector import AccessConfig
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.lancedb.lancedb import (
    LanceDBRemoteConnectionConfig,
    LanceDBUploader,
    LanceDBUploaderConfig,
    LanceDBUploadStager,
    LanceDBUploadStagerConfig,
)

CONNECTOR_TYPE = "lancedb_azure"


class LanceDBAzureAccessConfig(AccessConfig):
    azure_storage_account_name: str = Field(description="The name of the azure storage account.")
    azure_storage_account_key: str = Field(description="The serialized azure service account key.")


class LanceDBAzureConnectionConfig(LanceDBRemoteConnectionConfig):
    access_config: Secret[LanceDBAzureAccessConfig]

    def get_storage_options(self) -> dict:
        return {**self.access_config.get_secret_value().model_dump(), "timeout": self.timeout}


@dataclass
class LanceDBAzureUploader(LanceDBUploader):
    upload_config: LanceDBUploaderConfig
    connection_config: LanceDBAzureConnectionConfig
    connector_type: str = CONNECTOR_TYPE


lancedb_azure_destination_entry = DestinationRegistryEntry(
    connection_config=LanceDBAzureConnectionConfig,
    uploader=LanceDBAzureUploader,
    uploader_config=LanceDBUploaderConfig,
    upload_stager_config=LanceDBUploadStagerConfig,
    upload_stager=LanceDBUploadStager,
)



import asyncio
from dataclasses import dataclass
from typing import Callable, Optional

from unstructured_ingest.data_types.file_data import file_data_from_file
from unstructured_ingest.logger import logger
from unstructured_ingest.pipeline.interfaces import PipelineStep
from unstructured_ingest.processes.filter import Filterer

STEP_ID = "filter"


@dataclass
class FilterStep(PipelineStep):
    process: Filterer
    identifier: str = STEP_ID

    def __post_init__(self):
        config = self.process.config.model_dump_json() if self.process.config else None
        logger.info(f"created {self.identifier} with configs: {config}")

    async def _run_async(self, fn: Callable, file_data_path: str, **kwargs) -> Optional[dict]:
        file_data = file_data_from_file(path=file_data_path)
        fn_kwargs = {"file_data": file_data}
        if not asyncio.iscoroutinefunction(fn):
            resp = fn(**fn_kwargs)
        elif semaphore := self.context.semaphore:
            async with semaphore:
                resp = await fn(**fn_kwargs)
        else:
            resp = await fn(**fn_kwargs)

        if resp:
            return {"file_data_path": file_data_path}
        return None



from __future__ import annotations

from unstructured_ingest.processes.connector_registry import (
    add_destination_entry,
    add_source_entry,
)

from .cloud import CONNECTOR_TYPE as CLOUD_CONNECTOR
from .cloud import kafka_cloud_destination_entry, kafka_cloud_source_entry
from .local import CONNECTOR_TYPE as LOCAL_CONNECTOR
from .local import kafka_local_destination_entry, kafka_local_source_entry

add_source_entry(source_type=LOCAL_CONNECTOR, entry=kafka_local_source_entry)
add_destination_entry(destination_type=LOCAL_CONNECTOR, entry=kafka_local_destination_entry)

add_source_entry(source_type=CLOUD_CONNECTOR, entry=kafka_cloud_source_entry)
add_destination_entry(destination_type=CLOUD_CONNECTOR, entry=kafka_cloud_destination_entry)



import logging
from dataclasses import dataclass

import click

from unstructured_ingest.cli.base.cmd import BaseCmd
from unstructured_ingest.cli.utils.click import Dict, conform_click_options
from unstructured_ingest.cli.utils.model_conversion import options_from_base_model
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry


@dataclass
class DestCmd(BaseCmd):
    registry_entry: DestinationRegistryEntry

    def get_registry_options(self):
        options = []
        configs = [
            config
            for config in [
                self.registry_entry.uploader_config,
                self.registry_entry.upload_stager_config,
                self.registry_entry.connection_config,
            ]
            if config
        ]
        for config in configs:
            options.extend(options_from_base_model(model=config))
        options = self.consolidate_options(options=options)
        return options

    def cmd(self, ctx: click.Context, **options) -> None:
        logger.setLevel(logging.DEBUG if options.get("verbose", False) else logging.INFO)
        if not ctx.parent:
            raise click.ClickException("destination command called without a parent")
        if not ctx.parent.info_name:
            raise click.ClickException("parent command missing info name")
        source_cmd = ctx.parent.info_name.replace("-", "_")
        source_options: dict = ctx.parent.params if ctx.parent else {}
        conform_click_options(options)
        try:
            pipeline = self.get_pipeline(
                src=source_cmd,
                source_options=source_options,
                dest=self.cmd_name,
                destination_options=options,
            )
            pipeline.run()
        except Exception as e:
            logger.error(f"failed to run destination command {self.cmd_name}: {e}", exc_info=True)
            raise click.ClickException(str(e)) from e

    def get_cmd(self) -> click.Command:
        # Dynamically create the command without the use of click decorators
        fn = self.cmd
        fn = click.pass_context(fn)
        cmd = click.command(fn)
        if not isinstance(cmd, click.core.Command):
            raise ValueError(f"generated command was not of expected type Command: {type(cmd)}")
        cmd.name = self.cli_cmd_name
        cmd.invoke_without_command = True
        self.add_options(cmd)
        cmd.params.append(
            click.Option(
                ["--custom-stager"],
                required=False,
                type=str,
                default=None,
                help="Pass a pointer to a custom upload stager to use, "
                "must be in format '<module>:<attribute>'",
            )
        )
        cmd.params.append(
            click.Option(
                ["--custom-stager-config-kwargs"],
                required=False,
                type=Dict(),
                default=None,
                help="Any kwargs to instantiate the configuration "
                "associated with the customer stager",
            )
        )
        return cmd



import contextlib
from contextlib import contextmanager
from dataclasses import dataclass, field
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import (
    FileDataSourceMetadata,
)
from unstructured_ingest.errors_v2 import ProviderError, UserAuthError, UserError
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)
from unstructured_ingest.processes.utils.blob_storage import (
    BlobStoreUploadStager,
    BlobStoreUploadStagerConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

CONNECTOR_TYPE = "s3"

# https://docs.aws.amazon.com/AmazonS3/latest/userguide/object-keys.html#object-key-guidelines-avoid-characters
CHARACTERS_TO_AVOID = ["\\", "{", "^", "}", "%", "`", "]", '"', ">", "[", "~", "<", "#", "|"]

if TYPE_CHECKING:
    from s3fs import S3FileSystem


class S3IndexerConfig(FsspecIndexerConfig):
    pass


class S3AccessConfig(FsspecAccessConfig):
    key: Optional[str] = Field(
        default=None,
        description="If not anonymous, use this access key ID, if specified. Takes precedence "
        "over `aws_access_key_id` in client_kwargs.",
    )
    secret: Optional[str] = Field(
        default=None, description="If not anonymous, use this secret access key, if specified."
    )
    token: Optional[str] = Field(
        default=None, description="If not anonymous, use this security token, if specified."
    )


class S3ConnectionConfig(FsspecConnectionConfig):
    supported_protocols: list[str] = field(default_factory=lambda: ["s3", "s3a"], init=False)
    access_config: Secret[S3AccessConfig] = Field(default=S3AccessConfig(), validate_default=True)
    endpoint_url: Optional[str] = Field(
        default=None,
        description="Use this endpoint_url, if specified. Needed for "
        "connecting to non-AWS S3 buckets.",
    )
    anonymous: bool = Field(
        default=False, description="Connect to s3 without local AWS credentials."
    )
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)

    def get_access_config(self) -> dict[str, Any]:
        access_configs: dict[str, Any] = {"anon": self.anonymous}
        if self.endpoint_url:
            access_configs["endpoint_url"] = self.endpoint_url

        # Avoid injecting None by filtering out k,v pairs where the value is None
        access_configs.update(
            {k: v for k, v in self.access_config.get_secret_value().model_dump().items() if v}
        )
        return access_configs

    @requires_dependencies(["s3fs", "fsspec"], extras="s3")
    @contextmanager
    def get_client(self, protocol: str) -> Generator["S3FileSystem", None, None]:
        with super().get_client(protocol=protocol) as client:
            yield client

    def wrap_error(self, e: Exception) -> Exception:
        # s3fs maps botocore errors into python ones using mapping here:
        # https://github.com/fsspec/s3fs/blob/main/s3fs/errors.py
        if isinstance(e, PermissionError):
            return UserAuthError(e)
        if isinstance(e, FileNotFoundError):
            return UserError(f"File not found: {e}")
        if cause := getattr(e, "__cause__", None):
            error_response = cause.response
            error_meta = error_response["ResponseMetadata"]
            http_code = error_meta["HTTPStatusCode"]
            message = error_response["Error"].get("Message", str(e))
            if 400 <= http_code < 500:
                return UserError(message)
            if http_code >= 500:
                return ProviderError(message)
        logger.error(f"unhandled exception from s3 ({type(e)}): {e}", exc_info=True)
        return e


@dataclass
class S3Indexer(FsspecIndexer):
    connection_config: S3ConnectionConfig
    index_config: S3IndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def wrap_error(self, e: Exception) -> Exception:
        return self.connection_config.wrap_error(e=e)

    def get_path(self, file_info: dict) -> str:
        return file_info["Key"]

    def get_metadata(self, file_info: dict) -> FileDataSourceMetadata:
        path = file_info["Key"]
        date_created = None
        date_modified = None
        modified = file_info.get("LastModified")
        if modified:
            date_created = str(modified.timestamp())
            date_modified = str(modified.timestamp())

        file_size = file_info.get("size") if "size" in file_info else None
        file_size = file_size or file_info.get("Size")

        version = file_info.get("ETag").rstrip('"').lstrip('"') if "ETag" in file_info else None
        metadata: dict[str, str] = {}
        with (
            contextlib.suppress(AttributeError),
            self.connection_config.get_client(protocol=self.index_config.protocol) as client,
        ):
            metadata = client.metadata(path=path)
        record_locator = {
            "protocol": self.index_config.protocol,
            "remote_file_path": self.index_config.remote_url,
        }
        if metadata:
            record_locator["metadata"] = metadata
        issue_characters = [char for char in CHARACTERS_TO_AVOID if char in path]
        if issue_characters:
            logger.warning(
                f"File path {path} contains characters "
                f"that can cause issues with S3: {issue_characters}"
            )
        return FileDataSourceMetadata(
            date_created=date_created,
            date_modified=date_modified,
            date_processed=str(time()),
            version=version,
            url=f"{self.index_config.protocol}://{path}",
            record_locator=record_locator,
            filesize_bytes=file_size,
        )


class S3DownloaderConfig(FsspecDownloaderConfig):
    pass


@dataclass
class S3Downloader(FsspecDownloader):
    protocol: str = "s3"
    connection_config: S3ConnectionConfig
    connector_type: str = CONNECTOR_TYPE
    download_config: Optional[S3DownloaderConfig] = field(default_factory=S3DownloaderConfig)


class S3UploaderConfig(FsspecUploaderConfig):
    pass


@dataclass
class S3Uploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: S3ConnectionConfig
    upload_config: S3UploaderConfig = field(default=None)


s3_source_entry = SourceRegistryEntry(
    indexer=S3Indexer,
    indexer_config=S3IndexerConfig,
    downloader=S3Downloader,
    downloader_config=S3DownloaderConfig,
    connection_config=S3ConnectionConfig,
)

s3_destination_entry = DestinationRegistryEntry(
    uploader=S3Uploader,
    uploader_config=S3UploaderConfig,
    connection_config=S3ConnectionConfig,
    upload_stager_config=BlobStoreUploadStagerConfig,
    upload_stager=BlobStoreUploadStager,
)



from __future__ import annotations

from dataclasses import dataclass, field
from time import time
from typing import TYPE_CHECKING, Any, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileDataSourceMetadata
from unstructured_ingest.errors_v2 import (
    ProviderError,
    UserAuthError,
    UserError,
)
from unstructured_ingest.errors_v2 import (
    RateLimitError as CustomRateLimitError,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import (
    DestinationRegistryEntry,
    SourceRegistryEntry,
)
from unstructured_ingest.processes.connectors.fsspec.fsspec import (
    FsspecAccessConfig,
    FsspecConnectionConfig,
    FsspecDownloader,
    FsspecDownloaderConfig,
    FsspecIndexer,
    FsspecIndexerConfig,
    FsspecUploader,
    FsspecUploaderConfig,
)
from unstructured_ingest.processes.utils.blob_storage import (
    BlobStoreUploadStager,
    BlobStoreUploadStagerConfig,
)
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    pass

CONNECTOR_TYPE = "dropbox"


class DropboxIndexerConfig(FsspecIndexerConfig):
    def model_post_init(self, __context):
        if not self.path_without_protocol.startswith("/"):
            self.path_without_protocol = "/" + self.path_without_protocol


class DropboxAccessConfig(FsspecAccessConfig):
    token: Optional[str] = Field(
        default=None, description="Dropbox access token."
    )  # This is the short lived (4h) token that needs to be generated anew each time.
    app_key: Optional[str] = Field(default=None, description="Dropbox app key.")
    app_secret: Optional[str] = Field(default=None, description="Dropbox app secret.")
    refresh_token: Optional[str] = Field(
        default=None, description="Dropbox refresh token."
    )  # This is the long lived token that doesn't expire


class DropboxConnectionConfig(FsspecConnectionConfig):
    access_config: Secret[DropboxAccessConfig] = Field(
        default=DropboxAccessConfig(), validate_default=True
    )
    connector_type: str = Field(default=CONNECTOR_TYPE)

    @requires_dependencies(["dropbox"])
    def get_dropbox_access_token_from_refresh(
        self,
        refresh_token: str,
        app_key: str,
        app_secret: str,
    ) -> str:
        """
        Uses the Dropbox Python SDK to exchange a long-lived refresh token for an access token.
        """
        import dropbox

        dbx = dropbox.Dropbox(
            oauth2_access_token=None,
            oauth2_refresh_token=refresh_token,
            app_key=app_key,
            app_secret=app_secret,
        )

        # This call fetches a new short-lived token and auto-updates dbx._oauth2_access_token
        dbx.check_and_refresh_access_token()
        short_lived_token = dbx._oauth2_access_token  # Private attr, but standard usage
        return short_lived_token

    def get_access_config(self) -> dict[str, Any]:
        """
        Overrides the parent FsspecConnectionConfig.get_access_config() to ensure
        that we always provide an access token if refresh credentials exist.
        """
        base_conf = super().get_access_config()

        refresh_token = base_conf.get("refresh_token")
        app_key = base_conf.get("app_key")
        app_secret = base_conf.get("app_secret")

        # Standard scenario - we have refresh a token and creds provided
        # which we're going to use to retrieve access token
        if refresh_token and app_key and app_secret:
            logger.debug("Attempting to generate access token from refresh token...")
            new_token = self.get_dropbox_access_token_from_refresh(
                refresh_token=refresh_token,
                app_key=app_key,
                app_secret=app_secret,
            )
            if not new_token:
                raise ValueError(
                    "Unable to retrieve an access token from Dropbox. "
                    "Please check that your refresh token, app key, and secret are valid."
                )
            base_conf["token"] = new_token
        elif not base_conf.get("token"):  # we might already have an access token from outside
            # We have neither an existing short?lived token nor refresh credentials
            raise ValueError(
                "No valid token or refresh_token with app credentials was found. "
                "Please check that your refresh token, app key, and secret are valid "
                "or provide a valid short-lived token"
            )

        return base_conf

    @requires_dependencies(["dropbox"])
    def wrap_error(self, e: Exception) -> Exception:
        from dropbox.exceptions import AuthError, HttpError, RateLimitError

        if not isinstance(e, HttpError):
            logger.error(f"Unhandled Dropbox exception: {repr(e)}", exc_info=True)
            return e

        if isinstance(e, AuthError):
            raise UserAuthError(e.error)
        elif isinstance(e, RateLimitError):
            return CustomRateLimitError(e.error)

        status_code = e.status_code
        if 400 <= status_code < 500:
            if body := getattr(e, "body", None):
                return UserError(body)
            else:
                return UserError(e.body)
        if status_code >= 500:
            if body := getattr(e, "body", None):
                return ProviderError(body)
            else:
                return ProviderError(e.body)

        logger.error(f"Unhandled Dropbox HttpError: {repr(e)}", exc_info=True)
        return e


@dataclass
class DropboxIndexer(FsspecIndexer):
    connection_config: DropboxConnectionConfig
    index_config: DropboxIndexerConfig
    connector_type: str = CONNECTOR_TYPE

    def get_path(self, file_info: dict) -> str:
        return file_info["name"]

    def get_metadata(self, file_info: dict) -> FileDataSourceMetadata:
        path = file_info["name"].lstrip("/")
        date_created = None
        date_modified = None
        server_modified = file_info.get("server_modified")
        client_modified = file_info.get("client_modified")
        if server_modified and client_modified and server_modified > client_modified:
            date_created = str(client_modified.timestamp())
            date_modified = str(server_modified.timestamp())
        elif server_modified and client_modified and server_modified < client_modified:
            date_created = str(server_modified.timestamp())
            date_modified = str(client_modified.timestamp())

        file_size = file_info.get("size") if "size" in file_info else None

        version = file_info.get("content_hash")
        record_locator = {
            "protocol": self.index_config.protocol,
            "remote_file_path": self.index_config.remote_url,
            "file_id": file_info.get("id"),
        }
        return FileDataSourceMetadata(
            date_created=date_created,
            date_modified=date_modified,
            date_processed=str(time()),
            version=version,
            url=f"{self.index_config.protocol}://{path}",
            record_locator=record_locator,
            filesize_bytes=file_size,
        )


class DropboxDownloaderConfig(FsspecDownloaderConfig):
    pass


@dataclass
class DropboxDownloader(FsspecDownloader):
    protocol: str = "dropbox"
    connection_config: DropboxConnectionConfig
    connector_type: str = CONNECTOR_TYPE
    download_config: Optional[DropboxDownloaderConfig] = field(
        default_factory=DropboxDownloaderConfig
    )


class DropboxUploaderConfig(FsspecUploaderConfig):
    pass


@dataclass
class DropboxUploader(FsspecUploader):
    connector_type: str = CONNECTOR_TYPE
    connection_config: DropboxConnectionConfig
    upload_config: DropboxUploaderConfig = field(default=None)


dropbox_source_entry = SourceRegistryEntry(
    indexer=DropboxIndexer,
    indexer_config=DropboxIndexerConfig,
    downloader=DropboxDownloader,
    downloader_config=DropboxDownloaderConfig,
    connection_config=DropboxConnectionConfig,
)

dropbox_destination_entry = DestinationRegistryEntry(
    uploader=DropboxUploader,
    uploader_config=DropboxUploaderConfig,
    connection_config=DropboxConnectionConfig,
    upload_stager_config=BlobStoreUploadStagerConfig,
    upload_stager=BlobStoreUploadStager,
)



from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Optional

from pydantic import Field, Secret

from unstructured_ingest.data_types.file_data import FileData
from unstructured_ingest.error import DestinationConnectionError
from unstructured_ingest.interfaces import (
    AccessConfig,
    ConnectionConfig,
    Uploader,
    UploaderConfig,
    UploadStagerConfig,
)
from unstructured_ingest.logger import logger
from unstructured_ingest.processes.connector_registry import DestinationRegistryEntry
from unstructured_ingest.processes.connectors.duckdb.base import BaseDuckDBUploadStager
from unstructured_ingest.utils.data_prep import get_data_df
from unstructured_ingest.utils.dep_check import requires_dependencies

if TYPE_CHECKING:
    from duckdb import DuckDBPyConnection as DuckDBConnection
    from pandas import DataFrame

CONNECTOR_TYPE = "duckdb"


class DuckDBAccessConfig(AccessConfig):
    pass


class DuckDBConnectionConfig(ConnectionConfig):
    connector_type: str = Field(default=CONNECTOR_TYPE, init=False)
    database: Optional[str] = Field(
        default=None,
        description="Database name. Path to the DuckDB .db file. If the file does "
        "not exist, it will be created at the specified path.",
    )
    db_schema: Optional[str] = Field(
        default="main",
        description="Schema name. Schema in the database where the elements table is located.",
    )
    table: Optional[str] = Field(
        default="elements",
        description="Table name. Table name into which the elements data is inserted.",
    )
    access_config: Secret[DuckDBAccessConfig] = Field(
        default=DuckDBAccessConfig(), validate_default=True
    )

    def __post_init__(self):
        if self.database is None:
            raise ValueError(
                "A DuckDB connection requires a path to a *.db or *.duckdb file "
                "through the `database` argument"
            )

    @requires_dependencies(["duckdb"], extras="duckdb")
    @contextmanager
    def get_client(self) -> Generator["DuckDBConnection", None, None]:
        import duckdb

        with duckdb.connect(self.database) as client:
            yield client

    @contextmanager
    def get_cursor(self) -> Generator["DuckDBConnection", None, None]:
        with self.get_client() as client, client.cursor() as cursor:
            yield cursor


class DuckDBUploadStagerConfig(UploadStagerConfig):
    pass


@dataclass
class DuckDBUploadStager(BaseDuckDBUploadStager):
    upload_stager_config: DuckDBUploadStagerConfig = field(
        default_factory=lambda: DuckDBUploadStagerConfig()
    )


class DuckDBUploaderConfig(UploaderConfig):
    batch_size: int = Field(default=50, description="[Not-used] Number of records per batch")


@dataclass
class DuckDBUploader(Uploader):
    connector_type: str = CONNECTOR_TYPE
    upload_config: DuckDBUploaderConfig
    connection_config: DuckDBConnectionConfig

    def precheck(self) -> None:
        try:
            with self.connection_config.get_cursor() as cursor:
                cursor.execute("SELECT 1;")
        except Exception as e:
            logger.error(f"failed to validate connection: {e}", exc_info=True)
            raise DestinationConnectionError(f"failed to validate connection: {e}")

    def upload_dataframe(self, df: "DataFrame") -> None:
        logger.debug(f"uploading {len(df)} entries to {self.connection_config.database} ")

        with self.connection_config.get_client() as conn:
            conn.query(
                f"INSERT INTO {self.connection_config.db_schema}.{self.connection_config.table} BY NAME SELECT * FROM df"  # noqa: E501
            )

    @requires_dependencies(["pandas"], extras="duckdb")
    def run_data(self, data: list[dict], file_data: FileData, **kwargs: Any) -> None:
        import pandas as pd

        df = pd.DataFrame(data=data)
        self.upload_dataframe(df=df)

    @requires_dependencies(["pandas"], extras="duckdb")
    def run(self, path: Path, file_data: FileData, **kwargs: Any) -> None:
        df = get_data_df(path)
        self.upload_dataframe(df=df)


duckdb_destination_entry = DestinationRegistryEntry(
    connection_config=DuckDBConnectionConfig,
    uploader=DuckDBUploader,
    uploader_config=DuckDBUploaderConfig,
    upload_stager=DuckDBUploadStager,
    upload_stager_config=DuckDBUploadStagerConfig,
)



# https://developers.notion.com/reference/property-object#files
from dataclasses import dataclass, field
from typing import List, Optional

from htmlBuilder.tags import Div, HtmlTag

from unstructured_ingest.processes.connectors.notion.interfaces import DBCellBase, DBPropertyBase
from unstructured_ingest.processes.connectors.notion.types.file import FileObject


@dataclass
class Files(DBPropertyBase):
    id: str
    name: str
    type: str = "files"
    description: Optional[str] = None
    files: dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)


@dataclass
class FilesCell(DBCellBase):
    id: str
    files: List[FileObject]
    type: str = "files"
    name: Optional[str] = None

    @classmethod
    def from_dict(cls, data: dict):
        return cls(files=[FileObject.from_dict(f) for f in data.pop("files", [])], **data)

    def get_html(self) -> Optional[HtmlTag]:
        if not self.files:
            return None
        return Div([], [f.get_html() for f in self.files])



from abc import ABC
from copy import copy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any
from uuid import NAMESPACE_DNS, uuid5

from pydantic import BaseModel

from unstructured_ingest.data_types.file_data import FileData, SourceIdentifiers
from unstructured_ingest.interfaces.process import BaseProcess
from unstructured_ingest.logger import logger
from unstructured_ingest.utils.compression import TAR_FILE_EXT, ZIP_FILE_EXT, uncompress_file


class UncompressConfig(BaseModel):
    pass


@dataclass
class Uncompressor(BaseProcess, ABC):
    config: UncompressConfig = field(default_factory=UncompressConfig)

    def is_async(self) -> bool:
        return True

    def run(self, file_data: FileData, **kwargs: Any) -> list[FileData]:
        local_filepath = Path(file_data.local_download_path)
        if local_filepath.suffix not in TAR_FILE_EXT + ZIP_FILE_EXT:
            return [file_data]
        new_path = uncompress_file(filename=str(local_filepath))
        new_files = [i for i in Path(new_path).rglob("*") if i.is_file()]
        responses = []
        logger.debug(
            "uncompressed {} files from original file {}: {}".format(
                len(new_files), local_filepath, ", ".join([str(f) for f in new_files])
            )
        )
        for f in new_files:
            new_file_data = copy(file_data)
            new_file_data.identifier = str(uuid5(NAMESPACE_DNS, str(f)))
            new_file_data.local_download_path = str(f.resolve())
            new_rel_download_path = str(f).replace(str(Path(local_filepath.parent)), "")[1:]
            new_file_data.source_identifiers = SourceIdentifiers(
                filename=f.name,
                fullpath=str(file_data.source_identifiers.fullpath).replace(
                    file_data.source_identifiers.filename, new_rel_download_path
                ),
                rel_path=(
                    str(file_data.source_identifiers.rel_path).replace(
                        file_data.source_identifiers.filename, new_rel_download_path
                    )
                    if file_data.source_identifiers.rel_path
                    else None
                ),
            )
            responses.append(new_file_data)
        return responses

    async def run_async(self, file_data: FileData, **kwargs: Any) -> list[FileData]:
        return self.run(file_data=file_data, **kwargs)
